import base64
from urllib import parse

from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.http import Http404
from django.views.generic import View
from django_sendfile import sendfile


class FileGetterView(LoginRequiredMixin, View):
    """
    This view will deliver all the files and check if the user is logged in. As the whole application is only
    available privately, we also need to protect our user generated content (images etc.)
    """

    http_method_names = ["get"]

    def get(self, request, file_info):
        obj, file = self.get_object_from_url(request, file_info)
        if not obj:
            raise Http404
        return sendfile(request, file.path)

    def get_object_from_url(self, request, url):
        """
        Resolve objects based on urls provided by "get_url_for_object".
        If any of this fails, a 404 should be raised in the view, but logging information for developers must be available.
        """
        try:
            (app, model, id, field, hash) = base64.urlsafe_b64decode(parse.unquote(url)).decode().split("|")
            model = ContentType.objects.get(app_label=app, model=model)
            obj = self.get_object(model.model_class(), id, request=request, field=field)
            return obj, getattr(obj, field)
        except Exception:
            return None, None

    def get_object(self, model_class, id, **kwargs):
        return model_class._default_manager.get(id=id)



from django.conf import settings
from rest_framework import exceptions
from rest_framework.authentication import TokenAuthentication


class WebhookTokenUser:
    """
    We don't want to attach the token to a specific user, but the Authentication class needs to return a user-like
    object (e.g. which has an "is_authenticated" method).
    """

    @property
    def is_simple_token_dummy(self):
        return True

    @property
    def is_authenticated(self):
        return True

    @property
    def is_anonymous(self):
        return False

    @property
    def is_superuser(self):
        return False

    def has_perm(self, *args, **kwargs):
        """
        This "user" does not have any special permissions.
        """
        return False


class WebhookTokenAuthentication(TokenAuthentication):
    def authenticate_credentials(self, key):
        if not settings.IANUS_DELETE_HOOK_API_TOKEN:
            raise exceptions.AuthenticationFailed("No token set on server.")

        if key != settings.IANUS_DELETE_HOOK_API_TOKEN:
            raise exceptions.AuthenticationFailed("Invalid token.")
        return WebhookTokenUser(), None



from django.core.management import BaseCommand
from sentry_sdk import capture_exception


class SentryTestException(Exception):
    pass


def bad():
    raise SentryTestException()


class Command(BaseCommand):
    """
    Use this command to test the connection to Sentry. The command raises an
    exception on purpose. This should be reported in Sentry. You may immediately
    resolve the issue in Sentry then.
    """

    def handle(self, *args, **options):
        # Manually generate a Sentry event as per https://docs.sentry.io/platforms/python/guides/django/#verify
        try:
            bad()
        except Exception as e:
            capture_exception(e)



import json
import logging

from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth import login
from django.core.management import call_command
from django.db import connection
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from django_utils.testing.filestructure import cleanup_path
from django_utils.testing.filestructure import filestructure_snapshot


logger = logging.getLogger(__name__)

SNAPSHOT_BEFORE = []


@method_decorator(csrf_exempt, name="dispatch")
class E2ETestSetupView(View):
    def reset_database(self):
        with connection.cursor() as cursor:
            cursor.execute(
                """
                DO
                $func$
                BEGIN
                EXECUTE
                (SELECT 'TRUNCATE TABLE ' || string_agg(oid::regclass::text, ', ') || ' RESTART IDENTITY'
                    FROM   pg_class
                    WHERE  relkind = 'r'  -- only tables
                    AND    relnamespace = 'public'::regnamespace
                );
                END
                $func$;
                """
            )

    def clear_solr(self):
        haystack_connections = getattr(settings, "HAYSTACK_CONNECTIONS", {})
        if not haystack_connections:
            return
        default_engine = haystack_connections.get("default", {}).get("ENGINE", "")
        if default_engine != "haystack.backends.solr_backend.SolrEngine":
            return
        call_command("clear_index", "--noinput")

    def load_initial_data(self):
        call_command("load_e2e_data", datasets=["initial"])

    def restore_permissions(self):
        """
        Restore permissions using a management command provided by django-extensions.
        """
        call_command("update_permissions")

    def post(self, request, *args, **kwargs):
        logger.info("E2E test setup")

        # Create a snapshot of the media root folder before the test
        SNAPSHOT_BEFORE.extend(filestructure_snapshot(settings.MEDIA_ROOT))

        self.reset_database()
        self.clear_solr()
        self.restore_permissions()
        self.load_initial_data()

        return HttpResponse(status=200)


@method_decorator(csrf_exempt, name="dispatch")
class E2ETestTearDownView(View):
    def post(self, request, *args, **kwargs):
        logger.info("E2E test teardown")

        # Cleanup the media root folder by removing any files added since the snapshot in the
        # `E2ETestSetupView`.
        cleanup_path(settings.MEDIA_ROOT, SNAPSHOT_BEFORE)
        SNAPSHOT_BEFORE.clear()

        return HttpResponse(status=200)


@method_decorator(csrf_exempt, name="dispatch")
class TestingLoginView(View):
    def post(self, request):
        data = json.loads(request.body)
        username = data.get("username", "username")
        password = data.get("password", "password")
        user = authenticate(request, username=username, password=password)
        if user and user.is_authenticated:
            login(request, user)
            return HttpResponse(status=200)
        return HttpResponse(status=401)


@method_decorator(csrf_exempt, name="dispatch")
class E2ETestLoadDataView(View):
    def post(self, request, *args, **kwargs):
        data = json.loads(request.body)
        datasets = data.get("datasets")
        logger.info(f"E2E loading datasets {datasets}")
        call_command("load_e2e_data", datasets=datasets)
        return HttpResponse(status=200)


class PingView(View):
    def head(self, request, *args, **kwargs):
        return HttpResponse("pong", status=200)



from rest_framework import serializers


class InquireDeleteUserSerializer(serializers.Serializer):
    userid = serializers.CharField()


class SubscribeUserDeleteSerializer(InquireDeleteUserSerializer):
    pass



import sentry_sdk
from configurations import values
from sentry_sdk.integrations.django import DjangoIntegration


class SentryMixin:
    SENTRY_DSN = values.Value(environ_required=True, environ_prefix="")

    # A path to an alternative CA bundle file in PEM-format.
    SENTRY_CACERTS = values.Value(None, environ_prefix="")

    SENTRY_TAGS = values.DictValue(default={}, environ_prefix="")
    SENTRY_ENVIRONMENT = values.Value("Production", environ_prefix="")
    SENTRY_SERVER_NAME = values.Value(None, environ_prefix="")
    SENTRY_IN_APP_INLCUDE = values.ListValue(default=[], environ_prefix="")
    SENTRY_IN_APP_EXCLUDE = values.ListValue(default=[], environ_prefix="")

    @classmethod
    def post_setup(cls):
        super().post_setup()
        cls.init_sentry_sdk()
        cls.configure_sentry_tags()

    @classmethod
    def init_sentry_sdk(cls):
        if cls.SENTRY_DSN is None:
            return

        sentry_sdk.init(**cls.make_sentry_init_kwargs())

    @classmethod
    def make_sentry_init_kwargs(cls):
        return dict(
            dsn=cls.SENTRY_DSN,
            integrations=[DjangoIntegration()],
            send_default_pii=True,
            environment=cls.get_sentry_environment(),
            ca_certs=cls.SENTRY_CACERTS,
            release=cls.get_sentry_release_version(),
            server_name=cls.SENTRY_SERVER_NAME,
            in_app_include=cls.SENTRY_IN_APP_INLCUDE,
            in_app_exclude=cls.SENTRY_IN_APP_EXCLUDE,
        )

    @classmethod
    def get_sentry_environment(cls):
        return cls.SENTRY_ENVIRONMENT

    @classmethod
    def get_sentry_release_version(cls):
        return None

    @classmethod
    def get_sentry_tags(cls):
        return cls.SENTRY_TAGS

    @classmethod
    def configure_sentry_tags(cls):
        if cls.SENTRY_DSN is None:
            return

        for tagname, tagvalue in cls.get_sentry_tags().items():
            sentry_sdk.set_tag(tagname, tagvalue)



from django.urls import path

from django_utils.testing import views


urlpatterns = [
    path("setup", views.E2ETestSetupView.as_view()),
    path("teardown", views.E2ETestTearDownView.as_view()),
    path("login", views.TestingLoginView.as_view()),
    path("load_data", views.E2ETestLoadDataView.as_view()),
    path("ping", views.PingView.as_view()),
]



from rest_framework.authentication import SessionAuthentication as DRFSessionAuthentication


class SessionAuthentication(DRFSessionAuthentication):
    """
    This class is needed because the default SessionAuthentication of "djangorestframework" does
    never return 401's, because they cannot fill the WWW-Authenticate header with a valid value
    in the 401 response. As a result, we cannot distinguish calls that are not unauthorized (401
    unauthorized) and calls for which the user does not have permission (403 forbidden).

    See https://github.com/encode/django-rest-framework/issues/5968 for the reasoning.

    We do set authenticate_header function in SessionAuthentication so that a value for the
    WWW-Authenticate header can be retrieved and the response code is automatically set to 401 in
    case of unauthenticated requests.
    """

    def authenticate_header(self, request):
        return "Session"



import sys

from configurations import values


class LogMixin:
    """
    This mixins contains the logging configuration of our Django based
    web applications. Its default values are suited for production.
    """

    LOGGING_LEVEL = values.Value("WARNING")

    LOGGING_HANDLERS = values.ListValue(["stream"])

    LOGGING_STREAM = sys.stderr

    @property
    def LOGGING_FILENAME(self):
        default_value = super().BASE_DIR / "log" / "django.log"
        return values.Value(environ_name="LOGGING_FILENAME", default=default_value)

    @property
    def LOGGING(self):
        return {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "verbose": {
                    "format": "{asctime} {levelname} {name} {message}",
                    "style": "{",
                },
                "simple": {
                    "format": "{message}",
                    "style": "{",
                },
            },
            "handlers": {
                "console": {
                    "level": "DEBUG",
                    "class": "logging.StreamHandler",
                    "formatter": "verbose",
                    "stream": self.LOGGING_STREAM,
                },
                "file": {
                    "level": "DEBUG",
                    "class": "logging.FileHandler",
                    "formatter": "verbose",
                    "filename": self.LOGGING_FILENAME,
                },
                "stream": {
                    "class": "logging.StreamHandler",
                    "formatter": "verbose",
                },
            },
            "loggers": {
                "": {
                    "handlers": self.LOGGING_HANDLERS,
                    "level": self.LOGGING_LEVEL,
                },
                "django": {
                    "handlers": self.LOGGING_HANDLERS,
                    "level": self.LOGGING_LEVEL,
                    "propagate": False,
                },
                "django.db": {
                    "handlers": self.LOGGING_HANDLERS,
                    "level": self.LOGGING_LEVEL,
                    "propagate": False,
                },
            },
        }



import importlib

from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework.response import Response
from rest_framework.views import APIView

from django_utils.webhooks.delete_user import default_delete_hooks
from django_utils.webhooks.delete_user.serializers import InquireDeleteUserSerializer
from django_utils.webhooks.delete_user.serializers import SubscribeUserDeleteSerializer
from django_utils.webhooks.delete_user.token_authentication import WebhookTokenAuthentication


User = get_user_model()


try:
    delete_hooks = importlib.import_module(settings.IANUS_USER_DELETE_HOOKS)
except ModuleNotFoundError:
    delete_hooks = default_delete_hooks


class InquireDeleteView(APIView):
    authentication_classes = [WebhookTokenAuthentication]

    def post(self, request, format=None):
        serializer = InquireDeleteUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            user = User.objects.get(username=serializer.validated_data["userid"])
            passed, reason = delete_hooks.inquire_delete(user)
            if not passed:
                return Response(data={"reason": reason}, status=412)
            return Response(status=204)
        except User.DoesNotExist:
            return Response(status=204)


class SubscribeDeleteView(APIView):
    authentication_classes = [WebhookTokenAuthentication]

    def post(self, request, format=None):
        serializer = SubscribeUserDeleteSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        try:
            user = User.objects.get(username=serializer.validated_data["userid"])
            delete_hooks.subscribe_delete(user)
            return Response(status=204)
        except User.DoesNotExist:
            return Response(status=204)



import base64
import hashlib
from urllib import parse

from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from rest_framework import serializers


class ProtectedImageField(serializers.ImageField):
    def get_url_for_object(self, obj, field_name):
        """
        Create a string with all information required to access the object, in the form of
        <app_label>|<model>|<pk>|<field_name>
        """

        hash = hashlib.sha256()
        if hasattr(obj, "modified"):
            hash.update(str(obj.modified).encode())

        content_type = ContentType.objects.get_for_model(obj)
        return parse.quote_plus(
            base64.urlsafe_b64encode(
                "|".join(
                    [
                        content_type.app_label,
                        content_type.model,
                        str(obj.pk),
                        field_name,
                        hash.hexdigest(),
                    ]
                ).encode()
            )
        )

    def to_representation(self, value):
        return reverse("media", kwargs={"file_info": self.get_url_for_object(value.instance, self.field_name)})



import os


def filestructure_snapshot(path):
    """
    Create a filestructure snapshot by returning the paths of all files inside a path.
    """
    paths = []
    for dirpath, _dirnames, filenames in os.walk(path):
        paths.append(dirpath)
        for filename in filenames:
            paths.append(os.path.join(dirpath, filename))
    return paths


def cleanup_path(path, snapshot_before):
    """
    Remove all files and directories inside a path which did not exist in the snapshot_before.
    """
    snapshot_after = filestructure_snapshot(path)
    to_delete = reversed(sorted(set(snapshot_after) - set(snapshot_before)))
    for path in to_delete:
        if os.path.isfile(path):
            os.unlink(path)
        if os.path.isdir(path):
            os.removedirs(path)



from django import forms
from django.db import models


class TextField(models.TextField):
    """
    A TextField (without max_length) which displays its form element as a single-line 'CharField'
    with extended width in the Django admin.
    """

    def formfield(self, **kwargs):
        formfield = super().formfield(**kwargs)
        if hasattr(formfield, "choices"):
            # Use the original widget (select) if choices are available.
            return formfield
        formfield.widget = forms.TextInput(attrs={"style": "width: 30em;"})
        return formfield



from django.urls import path

from django_utils.webhooks.delete_user import views


urlpatterns = [
    path("api/auth/delete_user/inquire/", views.InquireDeleteView.as_view()),
    path("api/auth/delete_user/delete/", views.SubscribeDeleteView.as_view()),
]
